import typing

from __feature__ import snake_case
from PySide6 import QtCore, QtGui, QtWidgets

from snakeribbon.constants import RibbonIcon
from snakeribbon.menu import RibbonMenu
from snakeribbon.tabbar import RibbonTabBar
from snakeribbon.utils import DataFile


class RibbonApplicationButton(QtWidgets.QToolButton):
    """Application button in the ribbon bar."""

    def add_file_menu(self) -> RibbonMenu:
        """Add a new ribbon menu to the application button.

        :return: The new ribbon menu.
        """
        menu = RibbonMenu(self)
        self.set_popup_mode(QtWidgets.QToolButton.ToolButtonPopupMode.InstantPopup)
        self.set_menu(menu)
        return menu


class RibbonTitleLabel(QtWidgets.QLabel):
    """Title label in the ribbon bar."""

    pass


class RibbonTitleWidget(QtWidgets.QFrame):
    """The title widget of the ribbon."""

    #: Signal, the help button was clicked.
    sig_help_button_clicked = QtCore.Signal(bool)

    #: Signal, the collapse button wa clicked.
    sig_collapse_ribbon_button_clicked = QtCore.Signal(bool)

    #: Buttons
    _quick_access_buttons = []
    _right_tool_buttons = []

    _quick_access_button_height = 20
    _right_button_height = 20

    # Mouse move events
    _start_point = None
    _window_point = None

    @typing.overload
    def __init__(self, title=None, parent=None):
        pass

    @typing.overload
    def __init__(self, parent=None):
        pass

    def __init__(self, *args, **kwargs):
        """Initialize the ribbon title widget.

        :param title: The title of the ribbon.
        :param parent: The parent widget.
        """
        if (args and not isinstance(args[0], QtWidgets.QWidget)) or ("title" in kwargs):
            title = args[0] if len(args) > 0 else kwargs.get("title", None)
            parent = args[1] if len(args) > 1 else kwargs.get("parent", None)
        else:
            title = "snakeribbon"
            parent = args[0] if len(args) > 0 else kwargs.get("parent", None)

        super().__init__(parent)
        # Tab bar layout
        self.set_size_policy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)  # type: ignore
        self._tab_bar_layout = QtWidgets.QHBoxLayout(self)
        self._tab_bar_layout.set_contents_margins(0, 0, 0, 0)
        self._tab_bar_layout.set_spacing(0)

        # Application
        self._application_button = RibbonApplicationButton()  # type: ignore
        self._application_button.set_icon(DataFile(RibbonIcon.Application))
        self._application_button.set_icon_size(QtCore.QSize(self._quick_access_button_height, self._quick_access_button_height))
        self._application_button.set_text("snakeribbon")
        self._application_button.set_tool_tip("snakeribbon")

        self._quick_access_tool_bar = QtWidgets.QToolBar()
        self._quick_access_tool_bar.set_icon_size(QtCore.QSize(self._quick_access_button_height, self._quick_access_button_height))
        self._quick_access_tool_bar.set_orientation(QtCore.Qt.Orientation.Horizontal)
        self._quick_access_tool_bar.set_movable(False)
        self._quick_access_tool_bar.add_widget(self._application_button)
        self._quick_access_tool_bar_widget = QtWidgets.QWidget()
        self._quick_access_tool_bar_layout = QtWidgets.QHBoxLayout(self._quick_access_tool_bar_widget)
        self._quick_access_tool_bar_layout.set_contents_margins(0, 0, 0, 0)
        self._quick_access_tool_bar_layout.set_spacing(0)
        self._quick_access_tool_bar_layout.add_widget(self._quick_access_tool_bar, 0, QtCore.Qt.AlignmentFlag.AlignBottom)

        # right toolbar
        self._right_tool_bar = QtWidgets.QToolBar()
        self._right_tool_bar.set_orientation(QtCore.Qt.Orientation.Horizontal)
        self._right_tool_bar.set_icon_size(QtCore.QSize(self._right_button_height, self._right_button_height))

        self._collapsable_ribbon_button = QtWidgets.QToolButton(self)
        self._collapsable_ribbon_button.set_icon_size(QtCore.QSize(self._right_button_height, self._right_button_height))
        self._collapsable_ribbon_button.set_icon(DataFile(RibbonIcon.Up))
        self._collapsable_ribbon_button.set_auto_raise(True)
        self._collapsable_ribbon_button.set_tool_tip("Collapse Ribbon")
        self._collapsable_ribbon_button.clicked.connect(self.sig_collapse_ribbon_button_clicked)  # type: ignore

        self._help_button = QtWidgets.QToolButton(self)
        self._help_button.set_icon_size(QtCore.QSize(self._right_button_height, self._right_button_height))
        self._help_button.set_icon(DataFile(RibbonIcon.Help))
        self._help_button.set_auto_raise(True)
        self._help_button.set_tool_tip("Help")
        self._help_button.clicked.connect(self.sig_help_button_clicked)  # type: ignore

        self.add_right_tool_button(self._collapsable_ribbon_button)
        self.add_right_tool_button(self._help_button)

        # category tab bar
        self._tab_bar = RibbonTabBar(self)
        self._tab_bar.set_expanding(False)
        self._tab_bar.set_size_policy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)  # type: ignore
        font = self._tab_bar.font()
        font.set_point_size(font.point_size() + 3)
        self._tab_bar.set_font(font)
        self._tab_bar.set_shape(QtWidgets.QTabBar.Shape.RoundedNorth)
        self._tab_bar.set_document_mode(True)

        # Title label
        self._title_label = RibbonTitleLabel(self)
        self._title_label.set_size_policy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)  # type: ignore
        self._title_label.set_alignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignBottom)  # type: ignore
        self._title_label.set_text(title)
        font = self._title_label.font()
        font.set_point_size(font.point_size() + 3)
        self._title_label.set_font(font)

        self._tab_bar_layout.add_widget(self._quick_access_tool_bar_widget, 0, QtCore.Qt.AlignmentFlag.AlignVCenter)
        self._tab_bar_layout.add_widget(self._tab_bar, 0, QtCore.Qt.AlignmentFlag.AlignVCenter)
        self._tab_bar_layout.add_widget(self._title_label, 1, QtCore.Qt.AlignmentFlag.AlignVCenter)
        self._tab_bar_layout.add_widget(self._right_tool_bar, 0, QtCore.Qt.AlignmentFlag.AlignVCenter)

    def application_button(self) -> RibbonApplicationButton:
        """Return the application button."""
        return self._application_button

    def set_application_icon(self, icon: QtGui.QIcon):
        """Set the application icon.

        :param icon: The icon to set.
        """
        self._application_button.set_icon(icon)

    def add_title_widget(self, widget: QtWidgets.QWidget):
        """Add a widget to the title layout.

        :param widget: The widget to add.
        """
        self._tab_bar_layout.add_widget(widget)

    def insert_title_widget(self, index: int, widget: QtWidgets.QWidget):
        """Insert a widget to the title layout.

        :param index: The index to insert the widget.
        :param widget: The widget to insert.
        """
        self._tab_bar_layout.insert_widget(index, widget)

    def remove_title_widget(self, widget: QtWidgets.QWidget):
        """Remove a widget from the title layout.

        :param widget: The widget to remove.
        """
        self._tab_bar_layout.remove_widget(widget)

    def tab_bar(self) -> RibbonTabBar:
        """Return the tab bar of the ribbon.

        :return: The tab bar of the ribbon.
        """
        return self._tab_bar

    def quick_access_tool_bar(self) -> QtWidgets.QToolBar:
        """Return the quick access toolbar of the ribbon.

        :return: The quick access toolbar of the ribbon.
        """
        return self._quick_access_tool_bar

    def quick_access_buttons(self) -> typing.List[QtWidgets.QToolButton]:
        """Return the quick access buttons of the ribbon.

        :return: The quick access buttons of the ribbon.
        """
        return self._quick_access_buttons

    def add_quick_access_button(self, button: QtWidgets.QToolButton):
        """Add a widget to the quick access bar.

        :param button: The button to add.
        """
        button.set_icon_size(QtCore.QSize(self._quick_access_button_height, self._quick_access_button_height))
        self._quick_access_buttons.append(button)
        self._quick_access_tool_bar.add_widget(button)

    def set_quick_access_button_height(self, height: int):
        """Set the height of the quick access buttons.

        :param height: The height to set.
        """
        self._quick_access_button_height = height
        self._application_button.set_icon(self._application_button.icon().pixmap(height, height))
        self._quick_access_tool_bar.set_icon_size(QtCore.QSize(height, height))

    def title(self) -> str:
        """Return the title of the ribbon.

        :return: The title of the ribbon.
        """
        return self._title_label.text()

    def set_title(self, title: str):
        """Set the title of the ribbon.

        :param title: The title to set.
        """
        self._title_label.set_text(title)

    def right_tool_bar(self) -> QtWidgets.QToolBar:
        """Return the right toolbar of the ribbon.

        :return: The right toolbar of the ribbon.
        """
        return self._right_tool_bar

    def add_right_tool_button(self, button: QtWidgets.QToolButton):
        """Add a widget to the right button bar.

        :param button: The button to add.
        """
        button.set_icon_size(QtCore.QSize(self._right_button_height, self._right_button_height))
        self._right_tool_buttons.append(button)
        self._right_tool_bar.add_widget(button)

    def set_right_tool_bar_height(self, height: int):
        """Set the height of the right buttons.

        :param height: The height to set.
        """
        self._right_button_height = height
        self._right_tool_bar.set_icon_size(QtCore.QSize(height, height))

    def help_ribbon_button(self) -> QtWidgets.QToolButton:
        """Return the help ribbon button.

        :return: The help ribbon button.
        """
        return self._help_button

    def set_help_button_icon(self, icon: QtGui.QIcon):
        """Set the icon of the help button.

        :param icon: The icon to set.
        """
        self._help_button.set_icon(icon)

    def remove_help_button(self):
        """Remove the help button from the ribbon."""
        self._help_button.set_visible(False)

    def set_collapse_button_icon(self, icon: QtGui.QIcon):
        """Set the icon of the min button.

        :param icon: The icon to set.
        """
        self._collapsable_ribbon_button.set_icon(icon)

    def remove_collapse_button(self):
        """Remove the min button from the ribbon."""
        self._collapsable_ribbon_button.set_visible(False)

    def collapse_ribbon_button(self) -> QtWidgets.QToolButton:
        """Return the collapse ribbon button.

        :return: The collapse ribbon button.
        """
        return self._collapsable_ribbon_button

    def set_title_widget_height(self, height: int):
        """Set the height of the title widget.

        :param height: The height to set.
        """
        self.set_quick_access_button_height(height)
        self.set_right_tool_bar_height(height)

    def top_level_widget(self) -> QtWidgets.QWidget:
        widget = self
        while widget.parentWidget():
            widget = widget.parentWidget()
        return widget

    def mouse_press_event(self, e: QtGui.QMouseEvent):
        self._start_point = e.pos()
        self._window_point = self.top_level_widget().frame_geometry().top_left()

    def mouse_move_event(self, e: QtGui.QMouseEvent):
        relpos = e.pos() - self._start_point if self._start_point else None
        self.top_level_widget().move(self._window_point + relpos) if self._window_point and relpos else None
        self.top_level_widget().window_handle().start_system_move()

    def mouseDoubleClickEvent(self, e: QtGui.QMouseEvent):
        mainwindow = self.top_level_widget()
        mainwindow.show_normal() if mainwindow.is_maximized() else mainwindow.show_maximized()



from __feature__ import snake_case

from enum import IntEnum

from PySide6 import QtGui


class RibbonCategoryStyle(IntEnum):
    """The button style of a category."""

    Normal = 0
    Context = 1


Normal = RibbonCategoryStyle.Normal
Context = RibbonCategoryStyle.Context


#: A list of context category colors
context_colors = [
    QtGui.QColor(201, 89, 156),  # 玫红
    QtGui.QColor(242, 203, 29),  # 黄
    QtGui.QColor(255, 157, 0),  # 橙
    QtGui.QColor(14, 81, 167),  # 蓝
    QtGui.QColor(228, 0, 69),  # 红
    QtGui.QColor(67, 148, 0),  # 绿
]


class RibbonSpaceFindMode(IntEnum):
    """Mode to find available space in a grid layout, ColumnWise or RowWise."""

    ColumnWise = 0
    RowWise = 1


ColumnWise = RibbonSpaceFindMode.ColumnWise
RowWise = RibbonSpaceFindMode.RowWise


class RibbonButtonStyle(IntEnum):
    """Button style, Small, Medium, or Large."""

    Small = 0
    Medium = 1
    Large = 2


Small = RibbonButtonStyle.Small
Medium = RibbonButtonStyle.Medium
Large = RibbonButtonStyle.Large


class RibbonIcon:
    """
    Internal icons
    """
    Application = "icons/application.svg"
    Backward = "icons/backward.svg"
    Forward = "icons/forward.svg"
    Up = "icons/up.svg"
    Down = "icons/down.svg"
    More = "icons/more.svg"
    PanelOptions = "icons/linking.svg"
    Help = "icons/help.svg"
    Undo = "icons/undo.svg"
    Redo = "icons/redo.svg"



from pathlib import Path

from PySide6 import QtCore, QtGui, QtWidgets
from __feature__ import snake_case

import typing

from snakeribbon.category import (
    RibbonCategory,
    RibbonContextCategories,
    RibbonContextCategory,
    RibbonNormalCategory,
)
from snakeribbon.utils import DataFile
from snakeribbon.menu import RibbonMenu
from snakeribbon.tabbar import RibbonTabBar
from snakeribbon.constants import RibbonCategoryStyle, context_colors, RibbonIcon
from snakeribbon.titlewidget import RibbonApplicationButton, RibbonTitleWidget


class RibbonStackedWidget(QtWidgets.QStackedWidget):
    """Stacked widget that is used to display the ribbon."""

    def __init__(self, parent=None):
        """Create a new ribbon stacked widget.

        :param parent: The parent widget.
        """
        super().__init__(parent)
        effect = QtWidgets.QGraphicsDropShadowEffect()
        effect.set_offset(2, 2)
        self.set_graphics_effect(effect)


class RibbonBar(QtWidgets.QMenuBar):
    """The RibbonBar class is the top level widget that contains the ribbon."""

    #: Signal, the help button was clicked.
    sig_help_button_clicked = QtCore.Signal(bool)

    #: hide the ribbon bar automatically when the mouse press outside the ribbon bar
    _auto_hide_ribbon = False

    #: The categories of the ribbon.
    _categories: typing.Dict[str, RibbonCategory] = {}
    _context_category_count = 0

    #: Maximum rows
    _max_rows = 6

    #: Whether the ribbon is visible.
    _ribbon_visible = True

    #: heights of the ribbon elements
    _ribbon_height = 150

    #: current tab index
    _current_tab_index = 0

    def __init__(self, title: str = "Ribbon Bar Title", max_rows=6, parent=None):
        """Create a new ribbon.

        :param title: The title of the ribbon.
        :param max_rows: The maximum number of rows.
        :param parent: The parent widget of the ribbon.
        """
        super().__init__(parent)
        self._categories = {}
        self._max_rows = max_rows
        self._ribbon_title = title

    def init(self):
        self.set_fixed_height(self._ribbon_height)

        self._title_widget = RibbonTitleWidget(self._ribbon_title, self)
        self._stacked_widget = RibbonStackedWidget(self)

        # Main layout
        self._main_layout = QtWidgets.QVBoxLayout(self)
        self._main_layout.set_contents_margins(0, 0, 0, 0)
        self._main_layout.set_spacing(0)
        self._main_layout.add_widget(self._title_widget, 0)
        self._main_layout.add_widget(self._stacked_widget, 1)
        self._main_layout.set_size_constraint(QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)

        # Connect signals
        self._title_widget.sig_help_button_clicked.connect(self.sig_help_button_clicked)
        self._title_widget.sig_collapse_ribbon_button_clicked.connect(self._collapse_button_clicked)
        self._title_widget.tab_bar().currentChanged.connect(self.show_category_by_index)  # type: ignore

    def auto_hide_ribbon(self) -> bool:
        """Return whether the ribbon bar is automatically hidden when the mouse is pressed outside the ribbon bar.

        :return: Whether the ribbon bar is automatically hidden.
        """
        return self._auto_hide_ribbon

    def set_auto_hide_ribbon(self, autoHide: bool):
        """Set whether the ribbon bar is automatically hidden when the mouse is pressed outside the ribbon bar.

        :param autoHide: Whether the ribbon bar is automatically hidden.
        """
        self._auto_hide_ribbon = autoHide

    def event_filter(self, a0: QtCore.QObject, a1: QtCore.QEvent) -> bool:
        if self._auto_hide_ribbon and a1.type() == QtCore.QEvent.Type.HoverMove:
            self.set_ribbon_visible(self.under_mouse())
        return super().event_filter(a0, a1)

    def action_at(self, QPoint):
        raise NotImplementedError("RibbonBar.action_at() is not implemented in the ribbon bar.")

    def action_geometry(self, QAction):
        raise NotImplementedError("RibbonBar.action_geometry() is not implemented in the ribbon bar.")

    def active_action(self):
        raise NotImplementedError("RibbonBar.active_action() is not implemented in the ribbon bar.")

    def add_menu(self, *__args):
        raise NotImplementedError("RibbonBar.add_menu() is not implemented in the ribbon bar.")

    def add_action(self, *__args):
        raise NotImplementedError("RibbonBar.add_action() is not implemented in the ribbon bar.")

    def add_separator(self):
        raise NotImplementedError("RibbonBar.add_separator() is not implemented in the ribbon bar.")

    def clear(self):
        raise NotImplementedError("RibbonBar.clear() is not implemented in the ribbon bar.")

    def corner_widget(self, corner=None, *args, **kwargs):
        raise NotImplementedError("RibbonBar.corner_widget() is not implemented in the ribbon bar.")

    def insert_menu(self, QAction, QMenu):
        raise NotImplementedError("RibbonBar.insert_menu() is not implemented in the ribbon bar.")

    def insert_separator(self, QAction):
        raise NotImplementedError("RibbonBar.insert_separator() is not implemented in the ribbon bar.")

    def is_default_up(self):
        raise NotImplementedError("RibbonBar.is_default_up() is not implemented in the ribbon bar.")

    def is_native_menu_bar(self):
        raise NotImplementedError("RibbonBar.is_native_menu_bar() is not implemented in the ribbon bar.")

    def set_active_action(self, QAction):
        raise NotImplementedError("RibbonBar.set_active_action() is not implemented in the ribbon bar.")

    def set_corner_widget(self, QWidget, corner=None, *args, **kwargs):
        raise NotImplementedError("RibbonBar.set_corner_widget() is not implemented in the ribbon bar.")

    def set_default_up(self, up):
        raise NotImplementedError("RibbonBar.set_default_up() is not implemented in the ribbon bar.")

    def set_native_menu_bar(self, bar):
        raise NotImplementedError("RibbonBar.set_native_menu_bar() is not implemented in the ribbon bar.")

    def application_option_button(self) -> RibbonApplicationButton:
        """Return the application button."""
        return self._title_widget.application_button()

    def set_application_text(self, text: str):
        return self._title_widget.application_button().set_text(text)

    def set_application_icon(self, icon: QtGui.QIcon):
        """Set the application icon.

        :param icon: The icon to set.
        """
        self._title_widget.application_button().set_icon(icon)

    def add_title_widget(self, widget: QtWidgets.QWidget):
        """Add a widget to the title widget.

        :param widget: The widget to add.
        """
        self._title_widget.add_title_widget(widget)

    def remove_title_widget(self, widget: QtWidgets.QWidget):
        """Remove a widget from the title widget.

        :param widget: The widget to remove.
        """
        self._title_widget.remove_title_widget(widget)

    def insert_title_widget(self, index: int, widget: QtWidgets.QWidget):
        """Insert a widget to the title widget.

        :param index: The index to insert the widget.
        :param widget: The widget to insert.
        """
        self._title_widget.insert_title_widget(index, widget)

    def add_file_menu(self) -> RibbonMenu:
        """Add a file menu to the ribbon."""
        return self.application_option_button().add_file_menu()

    def ribbon_height(self) -> int:
        """Get the total height of the ribbon.

        :return: The height of the ribbon.
        """
        return self._ribbon_height

    def set_ribbon_height(self, height: int):
        """Set the total height of the ribbon.

        :param height: The height to set.
        """
        self._ribbon_height = height
        self.set_fixed_height(height)

    def tab_bar(self) -> RibbonTabBar:
        """Return the tab bar of the ribbon.

        :return: The tab bar of the ribbon.
        """
        return self._title_widget.tab_bar()

    def quick_access_tool_bar(self) -> QtWidgets.QToolBar:
        """Return the quick access toolbar of the ribbon.

        :return: The quick access toolbar of the ribbon.
        """
        return self._title_widget.quick_access_tool_bar()

    def add_quick_access_button(self, button: QtWidgets.QToolButton):
        """Add a button to the quick access bar.

        :param button: The button to add.
        """
        button.set_auto_raise(True)
        self._title_widget.quick_access_tool_bar().add_widget(button)

    def set_quick_access_button_height(self, height: int):
        """Set the height of the quick access buttons.

        :param height: The height to set.
        """
        self._title_widget.set_quick_access_button_height(height)

    def title(self) -> str:
        """Return the title of the ribbon.

        :return: The title of the ribbon.
        """
        return self._title_widget.title()

    def set_title(self, title: str):
        """Set the title of the ribbon.

        :param title: The title to set.
        """
        self._title_widget.set_title(title)

    def set_title_widget_height(self, height: int):
        """Set the height of the title widget.

        :param height: The height to set.
        """
        self._title_widget.set_title_widget_height(height)

    def right_tool_bar(self) -> QtWidgets.QToolBar:
        """Return the right toolbar of the ribbon.

        :return: The right toolbar of the ribbon.
        """
        return self._title_widget.right_tool_bar()

    def add_right_tool_button(self, button: QtWidgets.QToolButton):
        """Add a widget to the right button bar.

        :param button: The button to add.
        """
        button.set_auto_raise(True)
        self._title_widget.add_right_tool_button(button)

    def set_right_tool_bar_height(self, height: int):
        """Set the height of the right buttons.

        :param height: The height to set.
        """
        self._title_widget.set_right_tool_bar_height(height)

    def help_ribbon_button(self) -> QtWidgets.QToolButton:
        """Return the help button of the ribbon.

        :return: The help button of the ribbon.
        """
        return self._title_widget.help_ribbon_button()

    def set_help_button_icon(self, icon: QtGui.QIcon):
        """Set the icon of the help button.

        :param icon: The icon to set.
        """
        self._title_widget.set_help_button_icon(icon)

    def remove_help_button(self):
        """Remove the help button from the ribbon."""
        self._title_widget.remove_help_button()

    def collapse_ribbon_button(self) -> QtWidgets.QToolButton:
        """Return the collapse ribbon button.

        :return: The collapse ribbon button.
        """
        return self._title_widget.collapse_ribbon_button()

    def set_collapse_button_icon(self, icon: QtGui.QIcon):
        """Set the icon of the min button.

        :param icon: The icon to set.
        """
        self._title_widget.set_collapse_button_icon(icon)

    def remove_collapse_button(self):
        """Remove the min button from the ribbon."""
        self._title_widget.remove_collapse_button()

    def category(self, name: str) -> RibbonCategory:
        """Return the category with the given name.

        :param name: The name of the category.
        :return: The category with the given name.
        """
        return self._categories[name]

    def categories(self) -> typing.Dict[str, RibbonCategory]:
        """Return a list of categories of the ribbon.

        :return: A dict of categories of the ribbon.
        """
        return self._categories

    def add_categories_by(
        self,
        data: typing.Dict[
            str,  # title of the category
            typing.Dict,  # data of the category
        ],
    ) -> typing.Dict[str, RibbonCategory]:
        """Add categories from a dict.

        :param data: The dict of categories. The dict is of the form:

            .. code-block:: python

                {
                    "category-title": {
                        "style": RibbonCategoryStyle.Normal,
                        "color": QtCore.Qt.red,
                        "panels": {
                            "panel-title": {
                                "showPanelOptionButton": True,
                                "widgets": {
                                    "widget-name": {
                                        "type": "Button",
                                        "args": (),
                                        "kwargs": {  # or "arguments" for backward compatibility
                                            "key1": "value1",
                                            "key2": "value2"
                                        }
                                    },
                                }
                            },
                        },
                    }
                }
        :return: A dict of categories of the ribbon.
        """
        categories = {}
        for title, category_data in data.items():
            style = category_data.get("style", RibbonCategoryStyle.Normal)
            color = category_data.get("color", None)
            categories[title] = self.add_category(title, style, color)
            categories[title].add_panels_by(category_data.get("panels", {}))
        return categories

    def add_category(
        self,
        title: str,
        style=RibbonCategoryStyle.Normal,
        color: QtGui.QColor = None,
    ) -> typing.Union[RibbonNormalCategory, RibbonContextCategory]:
        """Add a new category to the ribbon.

        :param title: The title of the category.
        :param style: The button style of the category.
        :param color: The color of the context category, only used if style is Context, if None, the default color
                      will be used.
        :return: The newly created category.
        """
        if title in self._categories:
            raise ValueError(f"Category with title {title} already exists.")
        if style == RibbonCategoryStyle.Context:
            if color is None:
                color = context_colors[self._context_category_count % len(context_colors)]
                self._context_category_count += 1
        category = (
            RibbonContextCategory(title, color, self)  # noqa
            if style == RibbonCategoryStyle.Context
            else RibbonNormalCategory(title, self)  # noqa
        )
        category.set_maximum_rows(self._max_rows)
        category.set_fixed_height(
            self._ribbon_height
            - self._main_layout.spacing() * 2
            - self._main_layout.contents_margins().top()
            - self._main_layout.contents_margins().bottom()
            - self._title_widget.height()
        )  # 4: extra space for drawing lines when debugging

        self._categories[title] = category
        self._stacked_widget.add_widget(category)

        if style == RibbonCategoryStyle.Normal:
            self._title_widget.tab_bar().add_tab(title, color)

        elif style == RibbonCategoryStyle.Context:
            category.hide()

        if len(self._categories) == 1:
            self._title_widget.tab_bar().set_current_index(1)
            self.show_category_by_index(1)

        return category

    def add_normal_category(self, title: str) -> RibbonNormalCategory:
        """Add a new category to the ribbon.

        :param title: The title of the category.
        :return: The newly created category.
        """
        return self.add_category(title, RibbonCategoryStyle.Normal)

    def add_context_category(
        self,
        title: str,
        color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor] = QtCore.Qt.GlobalColor.blue,
    ) -> RibbonContextCategory:
        """Add a new context category to the ribbon.

        :param title: The title of the category.
        :param color: The color of the context category, if None, the default color will be used.
        :return: The newly created category.
        """
        return self.add_category(title, RibbonCategoryStyle.Context, color)

    def add_context_categories(
        self,
        name: str,
        titles: typing.List[str],
        color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor] = QtCore.Qt.GlobalColor.blue,
    ) -> RibbonContextCategories:
        """Add a group of context categories with the same tab color to the ribbon.

        :param name: The name of the context categories.
        :param titles: The title of the category.
        :param color: The color of the context category, if None, the default color will be used.
        :return: The newly created category.
        """
        if color is None:
            color = context_colors[self._context_category_count % len(context_colors)]
            self._context_category_count += 1
        categories = RibbonContextCategories(
            name,
            color,
            {title: self.add_context_category(title, color) for title in titles},
            self,
        )
        return categories

    def show_category_by_index(self, index: int):
        """Show category by tab index

        :param index: tab index
        """
        self._current_tab_index = index
        title = self._title_widget.tab_bar().tab_text(index)  # 0 is the file tab
        if title in self._categories:
            self._stacked_widget.set_current_widget(self._categories[title])

    def show_context_category(self, category: typing.Union[RibbonContextCategory, RibbonContextCategories]):
        """Show the given category or categories, if it is not a context category, nothing happens.

        :param category: The category to show.
        """
        if isinstance(category, RibbonContextCategory):
            self._title_widget.tab_bar().add_tab(category.title(), category.color())
            self._title_widget.tab_bar().set_current_index(self._title_widget.tab_bar().count() - 1)
            self._stacked_widget.set_current_widget(category)
        elif isinstance(category, RibbonContextCategories):
            categories = category
            titles = list(categories.keys())
            self._title_widget.tab_bar().add_associated_tabs(categories.name(), titles, categories.color())
            self._title_widget.tab_bar().set_current_index(self._title_widget.tab_bar().count() - len(titles))
            self._stacked_widget.set_current_widget(categories[titles[0]])

    def hide_context_category(self, category: typing.Union[RibbonContextCategory, RibbonContextCategories]):
        """Hide the given category or categories, if it is not a context category, nothing happens.

        :param category: The category to hide.
        """
        if isinstance(category, RibbonContextCategory):
            self.tab_bar().remove_tab(self.tab_bar().index_of(category.title()))
        elif isinstance(category, RibbonContextCategories):
            categories = category
            for c in categories:
                self.tab_bar().remove_tab(self.tab_bar().index_of(c.title()))

    def category_visible(self, category: RibbonCategory) -> bool:
        """Return whether the category is shown.

        :param category: The category to check.

        :return: Whether the category is shown.
        """
        return category.title() in self._title_widget.tab_bar().tab_titles()

    def remove_category(self, category: RibbonCategory):
        """Remove a category from the ribbon.

        :param category: The category to remove.
        """
        self.tab_bar().remove_tab(self._title_widget.tab_bar().index_of(category.title()))
        self._stacked_widget.remove_widget(category)

    def remove_categories(self, categories: RibbonContextCategories):
        """Remove a list of categories from the ribbon.

        :param categories: The categories to remove.
        """
        for category in categories.values():
            self.remove_category(category)

    def set_current_category(self, category: RibbonCategory):
        """Set the current category.

        :param category: The category to set.
        """
        self._stacked_widget.set_current_widget(category)
        if category.title() in self._title_widget.tab_bar().tab_titles():
            self._title_widget.tab_bar().set_current_index(self._title_widget.tab_bar().index_of(category.title()))
        else:
            raise ValueError(
                f"Category {category.title()} is not in the ribbon, "
                f"please show the context category/categories first."
            )

    def current_category(self) -> RibbonCategory:
        """Return the current category.

        :return: The current category.
        """
        return self._categories[self._title_widget.tab_bar().tab_text(self._title_widget.tab_bar().current_index())]

    def minimum_size_hint(self) -> QtCore.QSize:
        """Return the minimum size hint of the widget.

        :return: The minimum size hint.
        """
        return QtCore.QSize(super().minimum_size_hint().width(), self._ribbon_height)

    def _collapse_button_clicked(self):
        self.tab_bar().currentChanged.connect(self.show_ribbon)  # type: ignore
        self.hide_ribbon() if self._stacked_widget.is_visible() else self.show_ribbon()

    def show_ribbon(self):
        """Show the ribbon."""
        if not self._ribbon_visible:
            self._ribbon_visible = True
            self.collapse_ribbon_button().set_tool_tip("Collapse Ribbon")
            self.collapse_ribbon_button().set_icon(DataFile(RibbonIcon.Up))
            self._stacked_widget.set_visible(True)
            self.set_fixed_size(self.size_hint())

    def hide_ribbon(self):
        """Hide the ribbon."""
        if self._ribbon_visible:
            self._ribbon_visible = False
            self.collapse_ribbon_button().set_tool_tip("Expand Ribbon")
            self.collapse_ribbon_button().set_icon(DataFile(RibbonIcon.Down))
            self._stacked_widget.set_visible(False)
            self.set_fixed_size(self.size_hint().width(), self._title_widget.size().height() + 5)  # type: ignore

    def ribbon_visible(self) -> bool:
        """Get the visibility of the ribbon.

        :return: True if the ribbon is visible, False otherwise.
        """
        return self._ribbon_visible

    def set_ribbon_visible(self, visible: bool):
        """Set the visibility of the ribbon.

        :param visible: True to show the ribbon, False to hide it.
        """
        self.show_ribbon() if visible else self.hide_ribbon()



from __feature__ import snake_case

from PySide6 import QtCore
from PySide6 import QtWidgets

from snakeribbon.menu import RibbonMenu
from snakeribbon.constants import RibbonButtonStyle


class RibbonToolButton(QtWidgets.QToolButton):
    """Tool button that is showed in the ribbon."""

    _button_style: RibbonButtonStyle
    _large_button_icon_size = 64
    _medium_button_icon_size = 48
    _small_button_icon_size = 32
    _maximum_icon_size = 64

    def __init__(self, parent=None):
        """Create a new ribbon tool button.

        :param parent: The parent widget.
        """
        super().__init__(parent)

        # Styles
        self.set_button_style(RibbonButtonStyle.Large)
        self.set_auto_raise(True)
        self.set_focus_policy(QtCore.Qt.FocusPolicy.NoFocus)

    def set_maximum_icon_size(self, size: int):
        """Set the maximum icon size of the button.

        :param size: The maximum icon size of the button.
        """
        self._maximum_icon_size = size
        self.set_button_style(self._button_style)

    def maximum_icon_size(self) -> int:
        """Get the maximum icon size of the button.

        :return: The maximum icon size of the button.
        """
        return self._maximum_icon_size

    def set_button_style(self, style: RibbonButtonStyle):
        """Set the button style of the button.

        :param style: The button style of the button.
        """
        self._button_style = style
        if style == RibbonButtonStyle.Small:
            height = self._small_button_icon_size
            height = min(height, self._maximum_icon_size)
            self.set_icon_size(QtCore.QSize(height, height))
            self.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            self.set_style_sheet(
                """
                RibbonToolButton::menu-indicator {
                    subcontrol-origin: padding;
                    subcontrol-position: right;
                    right: -5px;
                }
                """
            )
        elif style == RibbonButtonStyle.Medium:
            height = self._medium_button_icon_size
            height = min(height, self._maximum_icon_size)
            self.set_icon_size(QtCore.QSize(height, height))
            self.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            self.set_style_sheet(
                """
                RibbonToolButton::menu-indicator {
                    subcontrol-origin: padding;
                    subcontrol-position: right;
                    right: -5px;
                }
                """
            )
        elif style == RibbonButtonStyle.Large:
            height = self._large_button_icon_size
            height = min(height, self._maximum_icon_size)
            self.set_icon_size(QtCore.QSize(height, height))
            self.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
            self.set_style_sheet(
                """
                RibbonToolButton[popupMode="0"]::menu-indicator {
                    subcontrol-origin: padding;
                    subcontrol-position: bottom;
                    bottom: -5px;
                }
                RibbonToolButton[popupMode="2"]::menu-indicator {
                    subcontrol-origin: padding;
                    subcontrol-position: bottom;
                    bottom: -5px;
                }
                """
            )

    def button_style(self) -> RibbonButtonStyle:
        """Get the button style of the button.

        :return: The button style of the button.
        """
        return self._button_style

    def add_ribbon_menu(self) -> RibbonMenu:
        """Add a ribbon menu for the button.

        :return: The added ribbon menu.
        """
        menu = RibbonMenu()
        self.set_menu(menu)
        return menu



from __feature__ import snake_case

import os
from typing import Union, Any


class _RegistryConnector:

    def __init__(self, registry: Union[dict, "BaseRegistry"] = None):
        self._registry = registry or {}

    def data(self, *args, **kwargs) -> Any:
        return None

    def set_registry(self, registry: Union[dict, "BaseRegistry"] = None):
        self._registry = registry

    def update(self, *args, **kwargs):
        raise NotImplementedError(f"Method `{self.__qualname__}` must be implemented")


class _DataFileConnector(_RegistryConnector):

    def data(self, filename: str) -> Any:
        return self._registry.get(filename)

    def update(self, icons: dict[str, Union[str, os.PathLike]]):
        self._registry.update(**icons)

    def __call__(self, filename: str) -> Any:
        return self._registry.get(filename)


DataFile = _DataFileConnector()


class _ThemeFileConnector(_RegistryConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._style_sheet = ""

    def data(self) -> Any:
        return self._style_sheet

    def update(self, filename: Union[str, os.PathLike]):
        with open(filename, "r", encoding="utf-8") as output:
            self._style_sheet += output.read()


ThemeFile = _ThemeFileConnector()



import os
import sys
from pathlib import Path
from typing import Union

from PySide6 import QtGui, QtWidgets
from PySide6.QtGui import QIcon, QPixmap, QPainter, QColor

from __feature__ import snake_case

from snakeribbon.ribbonbar import RibbonBar
from snakeribbon.constants import Large, RibbonIcon
from snakeribbon.constants import RowWise
from snakeribbon.utils import DataFile, ThemeFile


def svg(file: os.PathLike, color: Union[str, list] = None) -> QIcon:
    if not file:
        return QIcon()

    pixmap = QPixmap(file)
    painter = QPainter(pixmap)
    painter.set_composition_mode(QPainter.CompositionMode.CompositionMode_SourceIn)

    painter.fill_rect(pixmap.rect(), QColor.from_string(color))
    painter.end()

    return QIcon(pixmap)


if __name__ == "__main__":
    root = Path(__file__).resolve().parent

    app = QtWidgets.QApplication(sys.argv)
    app.set_font(QtGui.QFont("Times New Roman", 8))

    window = QtWidgets.QMainWindow()
    window.set_window_title("Pie-Ribbon Menu Bar")
    window.set_window_icon(QIcon(str(root / "icons/word.png")))
    centralWidget = QtWidgets.QWidget()
    window.set_central_widget(centralWidget)
    layout = QtWidgets.QVBoxLayout(centralWidget)

    # Define Ribbon instance without rendering it
    ribbon = RibbonBar()

    # Set-up internal icons and their path
    DataFile.update({
        RibbonIcon.Application: QIcon(str(root / "icons/save.png")),
        RibbonIcon.Backward: svg(root / "icons/svg/backward.svg"),
        RibbonIcon.Forward: svg(root / "icons/svg/box.svg"),
        RibbonIcon.Up: svg(root / "icons/svg/arrow-up.svg"),
        RibbonIcon.Down: svg(root / "icons/svg/arrow-down.svg"),
        RibbonIcon.More: svg(root / "icons/svg/box.svg"),
        RibbonIcon.PanelOptions: svg(root / "icons/svg/album.svg"),
        RibbonIcon.Help: svg(root / "icons/svg/info.svg"),
        RibbonIcon.Undo: svg(root / "icons/svg/undo.svg"),
        RibbonIcon.Redo: svg(root / "icons/svg/redo.svg"),
    })
    ThemeFile.update("themes/base.qss")
    ThemeFile.update("themes/default.qss")

    ribbon.set_style_sheet(ThemeFile.data())

    # Initialize Ribbon
    ribbon.init()
    window.set_menu_bar(ribbon)

    layout.add_widget(QtWidgets.QTextEdit(), 1)
    ribbon.set_application_text("App")
    ribbon.set_application_icon(DataFile(RibbonIcon.Application))

    undo_button = QtWidgets.QToolButton()
    undo_button.set_auto_raise(True)
    undo_button.set_text("Button")
    undo_button.set_icon(DataFile(RibbonIcon.Undo))
    undo_button.set_tool_tip("Undo")
    ribbon.add_quick_access_button(undo_button)

    redo_button = QtWidgets.QToolButton()
    redo_button.set_auto_raise(True)
    redo_button.set_text("Button")
    redo_button.set_icon(DataFile(RibbonIcon.Redo))
    redo_button.set_tool_tip("Redo")
    ribbon.add_quick_access_button(redo_button)

    # Home category
    home_category = ribbon.add_category("Home")
    options_category = ribbon.add_category("Options")

    clipboard_panel = home_category.add_panel("Clipboard", show_panel_option_button=True)
    clipboard_panel.panel_option_clicked.connect(lambda: print("SEX?"))
    paste_button = clipboard_panel.add_large_button(
        "Paste",
        icon=QIcon(str(root / "icons/paste.png")),
        tooltip="Paste"
    )
    paste_button.add_action(QtGui.QAction(QIcon(str(root / "icons/paste-special.png")), "Paste Special"))
    paste_button.add_action(QtGui.QAction(QIcon(str(root / "icons/paste-as-text.png")), "Paste as Text"))
    clipboard_panel.add_small_button(
        "Cut",
        icon=QIcon(str(root / "icons/close.png")),
        show_text=False,
        tooltip="Cut"
    )
    clipboard_panel.add_small_button(
        "Copy",
        icon=QIcon(str(root / "icons/close.png")),
        show_text=False,
        tooltip="Copy"
    )
    clipboard_panel.add_small_button(
        "Painter",
        icon=QIcon(str(root / "icons/close.png")),
        show_text=False,
        tooltip="Format Painter"
    )

    font_panel = home_category.add_panel("Font", show_panel_option_button=True)
    font_panel.add_small_toggle_button(
        "Bold",
        icon=QIcon(str(root / "icons/bold.png")),
        show_text=False,
        tooltip="Bold"
    )
    font_panel.add_small_toggle_button(
        "Italic",
        icon=QIcon(str(root / "icons/italic.png")),
        show_text=False,
        tooltip="Italic"
    )
    font_panel.add_small_toggle_button(
        "Underline",
        icon=QIcon(str(root / "icons/underline.png")),
        show_text=False,
        tooltip="Underline"
    )
    font_panel.add_small_toggle_button(
        "Toggle Button",
        icon=QIcon(str(root / "icons/strikethrough.png")),
        show_text=False,
        tooltip="Toggle me!"
    )
    font_panel.add_horizontal_separator()
    font_panel.add_date_time_edit(row_span=Large, col_span=Large)
    font_size_combo_box = font_panel.add_list_widget(["8", "9", "10"], fixed_height=True, mode=RowWise)
    spin_box = font_panel.add_double_spin_box(fixed_height=True, mode=RowWise)

    window.show_maximized()
    window.show()
    sys.exit(app.exec())



from __feature__ import snake_case

import typing

from PySide6 import QtCore, QtGui, QtWidgets

from snakeribbon.constants import RibbonIcon
from snakeribbon.menu import RibbonPermanentMenu
from snakeribbon.separator import RibbonHorizontalSeparator
from snakeribbon.toolbutton import RibbonToolButton
from snakeribbon.utils import DataFile


class RibbonPopupWidget(QtWidgets.QFrame):
    """The popup widget for the gallery widget."""


class RibbonGalleryListWidget(QtWidgets.QListWidget):
    """Gallery list widget."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.set_view_mode(QtWidgets.QListWidget.ViewMode.IconMode)
        self.set_resize_mode(QtWidgets.QListWidget.ResizeMode.Adjust)
        self.set_vertical_scroll_mode(QtWidgets.QListWidget.ScrollMode.ScrollPerPixel)
        self.set_horizontal_scroll_bar_policy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.set_vertical_scroll_bar_policy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.set_icon_size(QtCore.QSize(64, 64))

    def resize_event(self, e: QtGui.QResizeEvent) -> None:
        """Resize the list widget."""
        super().resize_event(e)

    def scroll_to_next_row(self) -> None:
        """Scroll to the next row."""
        self.vertical_scroll_bar().set_value(self.vertical_scroll_bar().value() + self.vertical_scroll_bar().single_step())

    def scroll_to_previous_row(self) -> None:
        """Scroll to the previous row."""
        self.vertical_scroll_bar().set_value(self.vertical_scroll_bar().value() - self.vertical_scroll_bar().single_step())


class RibbonGalleryButton(QtWidgets.QToolButton):
    """Gallery button."""


class RibbonGalleryPopupListWidget(RibbonGalleryListWidget):
    """Gallery popup list widget."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.set_vertical_scroll_bar_policy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)


class RibbonGallery(QtWidgets.QFrame):
    """A widget that displays a gallery of buttons."""

    _popup_window_size = QtCore.QSize(500, 500)
    _buttons: typing.List[RibbonToolButton] = []
    _popup_buttons: typing.List[RibbonToolButton] = []
    _popup_hide_on_click = False

    @typing.overload
    def __init__(self, minimum_width=800, popup_hide_on_click=False, parent=None):
        pass

    @typing.overload
    def __init__(self, parent=None):
        pass

    def __init__(self, *args, **kwargs):
        """Create a gallery.

        :param minimum_width: minimum width of the gallery
        :param popup_hide_on_click: hide on click flag
        :param parent: parent widget
        """
        if (args and not isinstance(args[0], QtWidgets.QWidget)) or (
            "minimum_width" in kwargs or "popup_hide_on_click" in kwargs
        ):
            minimum_width = args[0] if len(args) > 0 else kwargs.get("minimum_width", 800)
            popup_hide_on_click = args[1] if len(args) > 1 else kwargs.get("popup_hide_on_click", False)
            parent = args[2] if len(args) > 2 else kwargs.get("parent", None)
        else:
            minimum_width = 800
            popup_hide_on_click = False
            parent = args[0] if len(args) > 0 else kwargs.get("parent", None)
        super().__init__(parent)
        self.set_minimum_width(minimum_width)
        self._popup_hide_on_click = popup_hide_on_click

        self._main_layout = QtWidgets.QHBoxLayout(self)
        self._main_layout.set_contents_margins(5, 5, 5, 5)
        self._main_layout.set_spacing(5)

        self._up_button = RibbonGalleryButton(self)
        self._up_button.set_icon(DataFile(RibbonIcon.Up))
        self._up_button.set_icon_size(QtCore.QSize(24, 24))
        self._up_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._up_button.set_auto_raise(True)
        self._down_button = RibbonGalleryButton(self)
        self._down_button.set_icon(DataFile(RibbonIcon.Down))
        self._down_button.set_icon_size(QtCore.QSize(24, 24))
        self._down_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._down_button.set_auto_raise(True)
        self._more_button = RibbonGalleryButton(self)
        self._more_button.set_icon(DataFile(RibbonIcon.More))
        self._more_button.set_icon_size(QtCore.QSize(24, 24))
        self._more_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._more_button.set_auto_raise(True)
        self._scroll_button_layout = QtWidgets.QVBoxLayout()
        self._scroll_button_layout.set_contents_margins(0, 0, 0, 0)
        self._scroll_button_layout.set_spacing(2)
        self._scroll_button_layout.add_widget(self._up_button)
        self._scroll_button_layout.add_widget(self._down_button)
        self._scroll_button_layout.add_widget(self._more_button)

        self._list_widget = RibbonGalleryListWidget()
        self._main_layout.add_widget(self._list_widget)
        self._main_layout.add_layout(self._scroll_button_layout)

        self._up_button.clicked.connect(self._list_widget.scroll_to_previous_row)  # type: ignore
        self._down_button.clicked.connect(self._list_widget.scroll_to_next_row)  # type: ignore

        self._popup_widget = RibbonPopupWidget()  # type: ignore
        self._popup_widget.set_font(QtWidgets.QApplication.instance().font())  # type: ignore
        self._popup_widget.set_window_flag(QtCore.Qt.WindowType.Popup)
        self._popup_layout = QtWidgets.QVBoxLayout(self._popup_widget)
        self._popup_layout.set_contents_margins(5, 5, 5, 5)
        self._popup_layout.set_spacing(2)

        self._popup_list_widget = RibbonGalleryPopupListWidget()
        self._popup_layout.add_widget(self._popup_list_widget)
        self._popup_layout.add_widget(RibbonHorizontalSeparator())

        self._popup_menu = RibbonPermanentMenu()
        self._popup_menu.set_size_policy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)  # type: ignore
        self._popup_menu.action_added.connect(self._handle_popup_action)
        self._popup_layout.add_widget(self._popup_menu)

        self._more_button.clicked.connect(self.show_popup)  # type: ignore

    def _handle_popup_action(self, action: QtGui.QAction) -> None:
        """Handle a popup action."""
        if isinstance(action, QtGui.QAction):
            action.triggered.connect(self.hide_popup_widget)  # type: ignore

    def resize_event(self, a0: QtGui.QResizeEvent) -> None:
        """Resize the gallery."""
        height = self.height() - self._main_layout.contents_margins().top() - self._main_layout.contents_margins().bottom()
        self._up_button.set_fixed_size(height // 4, height // 3)  # type: ignore
        self._down_button.set_fixed_size(height // 4, height // 3)  # type: ignore
        self._more_button.set_fixed_size(height // 4, height // 3)  # type: ignore
        super().resize_event(a0)

    def popup_menu(self) -> RibbonPermanentMenu:
        """Return the popup menu."""
        return self._popup_menu

    def show_popup(self):
        """Show the popup window"""
        self._popup_widget.move(self.map_to_global(self.geometry().top_left()))
        self._popup_widget.resize(
            QtCore.QSize(
                max(self.popup_window_size().width(), self.width()), max(self.popup_window_size().height(), self.height())
            )
        )
        self._popup_menu.set_fixed_height(
            self._popup_widget.width()
            - self._popup_layout.contents_margins().left()
            - self._popup_layout.contents_margins().right()
        )
        self._popup_widget.show()

    def hide_popup_widget(self):
        """Hide the popup window"""
        self._popup_widget.hide()

    def popup_window_size(self):
        """Return the size of the popup window

        :return: size of the popup window
        """
        return self._popup_window_size

    def set_popup_window_size(self, size: QtCore.QSize):
        """Set the size of the popup window

        :param size: size of the popup window
        """
        self._popup_window_size = size

    def set_selected_button(self):
        """Set the selected button"""
        button = self.sender()
        if isinstance(button, RibbonToolButton):
            row = self._popup_buttons.index(button)
            self._list_widget.scroll_to(
                self._list_widget.model().index(row, 0), QtWidgets.QAbstractItemView.ScrollHint.EnsureVisible
            )
            if self._buttons[row].is_checkable():
                self._buttons[row].set_checked(not self._buttons[row].is_checked())

    def _add_widget(self, widget: QtWidgets.QWidget):
        """Add a widget to the gallery

        :param widget: widget to add
        """
        item = QtWidgets.QListWidgetItem()
        item.set_size_hint(widget.size_hint())
        self._list_widget.set_spacing((self.height() - item.size_hint().height()) // 2)
        self._list_widget.add_item(item)
        self._list_widget.set_item_widget(item, widget)

    def _add_popup_widget(self, widget: QtWidgets.QWidget):
        """Add a widget to the popup gallery

        :param widget: widget to add
        """
        item = QtWidgets.QListWidgetItem()
        item.set_size_hint(widget.size_hint())
        self._popup_list_widget.set_spacing((self.height() - item.size_hint().height()) // 2)
        self._popup_list_widget.add_item(item)
        self._popup_list_widget.set_item_widget(item, widget)

    def set_popup_hide_on_click(self, popup_hide_on_click: bool):
        """Set the hide on click flag

        :param popup_hide_on_click: hide on click flag
        """
        self._popup_hide_on_click = popup_hide_on_click

    def add_button(
        self,
        text: str = None,
        icon: QtGui.QIcon = None,
        slot=None,
        shortcut=None,
        tooltip=None,
        statusTip=None,
        checkable=False,
    ) -> typing.Tuple[RibbonToolButton, RibbonToolButton]:
        """Add a button to the gallery

        :param text: text of the button
        :param icon: icon of the button
        :param slot: slot to call when the button is clicked
        :param shortcut: shortcut of the button
        :param tooltip: tooltip of the button
        :param statusTip: status tip of the button
        :param checkable: checkable flag of the button.
        :return: the button and the popup button added
        """
        button = RibbonToolButton(self)
        popup_button = RibbonToolButton(self._popup_widget)
        if text is not None:
            button.set_text(text)
            popup_button.set_text(text)
        if icon is not None:
            button.set_icon(icon)
            popup_button.set_icon(icon)
        if slot is not None:
            button.clicked.connect(slot)  # type: ignore
            popup_button.clicked.connect(slot)  # type: ignore
        if shortcut is not None:
            button.set_shortcut(shortcut)
            popup_button.set_shortcut(shortcut)
        if tooltip is not None:
            button.set_tool_tip(tooltip)
            popup_button.set_tool_tip(tooltip)
        if statusTip is not None:
            button.set_status_tip(statusTip)
            popup_button.set_status_tip(statusTip)
        if checkable:
            button.set_checkable(True)
            popup_button.set_checkable(True)
        self._buttons.append(button)
        self._popup_buttons.append(popup_button)
        button.clicked.connect(lambda checked: popup_button.set_checked(checked))  # type: ignore
        if self._popup_hide_on_click:
            popup_button.clicked.connect(self.hide_popup_widget)  # type: ignore
        popup_button.clicked.connect(self.set_selected_button)  # type: ignore

        if text is None:
            button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
            popup_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        else:
            button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
            popup_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self._add_widget(button)  # noqa
        self._add_popup_widget(popup_button)  # noqa
        return button, popup_button

    def addToggleButton(
        self,
        text: str = None,
        icon: QtGui.QIcon = None,
        slot=None,
        shortcut=None,
        tooltip=None,
        statusTip=None,
    ) -> typing.Tuple[RibbonToolButton, RibbonToolButton]:
        """Add a toggle button to the gallery

        :param text: text of the button
        :param icon: icon of the button
        :param slot: slot to call when the button is clicked
        :param shortcut: shortcut of the button
        :param tooltip: tooltip of the button
        :param statusTip: status tip of the button.
        :return: the button and the popup button added
        """
        return self.add_button(text, icon, slot, shortcut, tooltip, statusTip, True)



from __future__ import annotations
from __feature__ import snake_case

import re
import functools
from typing import Any, Callable, Dict, List, Union, overload

import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from snakeribbon.constants import ColumnWise, RibbonIcon
from snakeribbon.constants import Large
from snakeribbon.constants import Medium
from snakeribbon.constants import RibbonButtonStyle
from snakeribbon.constants import RibbonSpaceFindMode
from snakeribbon.constants import Small
from snakeribbon.gallery import RibbonGallery
from snakeribbon.separator import RibbonSeparator
from snakeribbon.toolbutton import RibbonToolButton
from snakeribbon.utils import DataFile


class RibbonPanelTitle(QtWidgets.QLabel):
    """Widget to display the title of a panel."""


class RibbonGridLayoutManager(object):
    """Grid Layout Manager."""

    def __init__(self, rows: int):
        """Create a new grid layout manager.

        :param rows: The number of rows in the grid layout.
        """
        self.rows = rows
        self.cells = np.ones((rows, 1), dtype=bool)

    def request_cells(self, row_span: int = 1, col_span: int = 1, mode: RibbonSpaceFindMode = ColumnWise):
        """Request a number of available cells from the grid.

        :param row_span: The number of rows the cell should span.
        :param col_span: The number of columns the cell should span.
        :param mode: The mode of the grid.
        :return: row, col, the row and column of the requested cell.
        """
        if row_span > self.rows:
            raise ValueError("row_span is too large")
        if mode == ColumnWise:
            for row in range(self.cells.shape[0] - row_span + 1):
                for col in range(self.cells.shape[1] - col_span + 1):
                    if self.cells[row : row + row_span, col : col + col_span].all():
                        self.cells[row : row + row_span, col: col + col_span] = False
                        return row, col
        else:
            for col in range(self.cells.shape[1]):
                if self.cells[0, col:].all():
                    if self.cells.shape[1] - col < col_span:
                        self.cells = np.append(
                            self.cells, np.ones((self.rows, col_span - (self.cells.shape[1] - col)), dtype=bool), axis=1
                        )
                    self.cells[0, col:] = False
                    return 0, col
        cols = self.cells.shape[1]
        col_span1 = col_span
        if self.cells[:, -1].all():
            cols -= 1
            col_span1 -= 1
        self.cells = np.append(self.cells, np.ones((self.rows, col_span1), dtype=bool), axis=1)
        self.cells[:row_span, cols: cols + col_span] = False
        return 0, cols


class RibbonPanelItemWidget(QtWidgets.QFrame):
    """Widget to display a panel item."""

    def __init__(self, parent=None):
        """Create a new panel item.

        :param parent: The parent widget.
        """
        super().__init__(parent)
        self.set_layout(QtWidgets.QVBoxLayout())
        self.layout().set_contents_margins(0, 0, 0, 0)
        self.layout().set_spacing(0)
        self.layout().set_alignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.layout().set_size_constraint(QtWidgets.QLayout.SizeConstraint.SetMaximumSize)
        self.set_size_policy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)  # type: ignore

    def add_widget(self, widget):
        """Add a widget to the panel item.

        :param widget: The widget to add.
        """
        self.layout().add_widget(widget)


class RibbonPanelOptionButton(QtWidgets.QToolButton):
    """Button to display the options of a panel."""


class RibbonPanel(QtWidgets.QFrame):
    """Panel in the ribbon category."""

    #: maximal number of rows
    _max_rows: int = 6
    #: rows for large widgets
    _large_rows: int = 6
    #: rows for medium widgets
    _medium_rows: int = 3
    #: rows for small widgets
    _small_rows: int = 2
    #: GridLayout manager to request available cells.
    _grid_layout_manager: RibbonGridLayoutManager
    #: whether to show the panel option button
    _show_panel_option_button: bool

    #: widgets that are added to the panel
    _widgets: List[QtWidgets.QWidget] = []

    # height of the title widget
    _title_height: int = 15

    # Panel options signal
    panel_option_clicked = QtCore.Signal(bool)

    @overload
    def __init__(self, title: str = "", max_rows: int = 6, show_panel_option_button=True, parent=None):
        pass

    @overload
    def __init__(self, parent=None):
        pass

    def __init__(self, *args, **kwargs):
        """Create a new panel.

        :param title: The title of the panel.
        :param max_rows: The maximal number of rows in the panel.
        :param show_panel_option_button: Whether to show the panel option button.
        :param parent: The parent widget.
        """
        if (args and not isinstance(args[0], QtWidgets.QWidget)) or ("title" in kwargs or "max_rows" in kwargs):
            title = args[0] if len(args) > 0 else kwargs.get("title", "")
            max_rows = args[1] if len(args) > 1 else kwargs.get("max_rows", 6)
            show_panel_option_button = args[2] if len(args) > 2 else kwargs.get("show_panel_option_button", True)
            parent = args[3] if len(args) > 3 else kwargs.get("parent", None)
        else:
            title = ""
            max_rows = 6
            show_panel_option_button = True
            parent = args[0] if len(args) > 0 else kwargs.get("parent", None)
        super().__init__(parent)
        self._max_rows = max_rows
        self._large_rows = max_rows
        self._medium_rows = max(round(max_rows / 2), 1)
        self._small_rows = max(round(max_rows / 3), 1)
        self._grid_layout_manager = RibbonGridLayoutManager(self._max_rows)
        self._widgets = []
        self._show_panel_option_button = show_panel_option_button

        # Main layout
        self._main_layout = QtWidgets.QVBoxLayout(self)
        self._main_layout.set_contents_margins(0, 0, 0, 0)
        self._main_layout.set_spacing(0)

        # Actions layout
        self._actions_layout = QtWidgets.QGridLayout()
        self._actions_layout.set_contents_margins(5, 5, 5, 5)
        self._actions_layout.set_spacing(0)
        self._main_layout.add_layout(self._actions_layout, 1)

        # Title layout
        self._title_widget = QtWidgets.QWidget()
        self._title_widget.set_fixed_height(self._title_height)
        self._title_layout = QtWidgets.QHBoxLayout(self._title_widget)
        self._title_layout.set_contents_margins(0, 0, 0, 0)
        self._title_layout.set_spacing(0)
        self._title_label = RibbonPanelTitle()  # type: ignore
        self._title_label.set_text(title)
        self._title_label.set_alignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self._title_layout.add_widget(self._title_label, 1)

        # Panel option button
        if show_panel_option_button:
            self._panel_option = RibbonPanelOptionButton()  # type: ignore
            self._panel_option.set_auto_raise(True)
            self._panel_option.set_icon(DataFile(RibbonIcon.PanelOptions))
            self._panel_option.set_icon_size(QtCore.QSize(self._title_height, self._title_height))
            self._panel_option.set_tool_tip("Panel options")
            self._panel_option.clicked.connect(self.panel_option_clicked)  # type: ignore
            self._title_layout.add_widget(self._panel_option, 0)

        self._main_layout.add_widget(self._title_widget, 0)

    def maximum_rows(self) -> int:
        """Return the maximal number of rows in the panel.

        :return: The maximal number of rows in the panel.
        """
        return self._max_rows

    def large_rows(self) -> int:
        """Return the number of span rows for large widgets.

        :return: The number of span rows for large widgets.
        """
        return self._large_rows

    def medium_rows(self) -> int:
        """Return the number of span rows for medium widgets.

        :return: The number of span rows for medium widgets.
        """
        return self._medium_rows

    def small_rows(self) -> int:
        """Return the number of span rows for small widgets.

        :return: The number of span rows for small widgets.
        """
        return self._small_rows

    def set_maximum_rows(self, max_rows: int):
        """Set the maximal number of rows in the panel.

        :param max_rows: The maximal number of rows in the panel.
        """
        self._max_rows = max_rows
        self._large_rows = max_rows
        self._medium_rows = max(round(max_rows / 2), 1)
        self._small_rows = max(round(max_rows / 3), 1)

    def set_large_rows(self, rows: int):
        """Set the number of span rows for large widgets.

        :param rows: The number of span rows for large widgets.
        """
        assert rows <= self._max_rows, "Invalid number of rows"
        self._large_rows = rows

    def set_medium_rows(self, rows: int):
        """Set the number of span rows for medium widgets.

        :param rows: The number of span rows for medium widgets.
        """
        assert 0 < rows <= self._max_rows, "Invalid number of rows"
        self._medium_rows = rows

    def set_small_rows(self, rows: int):
        """Set the number of span rows for small widgets.

        :param rows: The number of span rows for small widgets.
        """
        assert 0 < rows <= self._max_rows, "Invalid number of rows"
        self._small_rows = rows

    def default_row_span(self, row_span: Union[int, RibbonButtonStyle]) -> int:
        """Return the number of span rows for the given widget type.

        :param row_span: row span or type.
        :return: The number of span rows for the given widget type.
        """
        if not isinstance(row_span, RibbonButtonStyle):
            return row_span
        if row_span == Large:
            return self._large_rows
        elif row_span == Medium:
            return self._medium_rows
        elif row_span == Small:
            return self._small_rows
        else:
            raise ValueError("Invalid row span")

    def panel_option_button(self) -> RibbonPanelOptionButton:
        """Return the panel option button.

        :return: The panel option button.
        """
        return self._panel_option

    def set_panel_option_tool_tip(self, text: str):
        """Set the tooltip of the panel option button.

        :param text: The tooltip text.
        """
        self._panel_option.set_tool_tip(text)

    def row_height(self) -> int:
        """Return the height of a row."""
        return int(
            (
                    self.size().height()
                    - self._main_layout.contents_margins().top()
                    - self._main_layout.contents_margins().bottom()
                    - self._main_layout.spacing()
                    - self._title_widget.height()
                    - self._actions_layout.contents_margins().top()
                    - self._actions_layout.contents_margins().bottom()
                    - self._actions_layout.vertical_spacing() * (self._grid_layout_manager.rows - 1)
            )
            / self._grid_layout_manager.rows
        )

    def set_title(self, title: str):
        """Set the title of the panel.

        :param title: The title to set.
        """
        self._title_label.set_text(title)

    def title(self):
        """Get the title of the panel.

        :return: The title.
        """
        return self._title_label.text()

    def set_title_height(self, height: int):
        """Set the height of the title widget.

        :param height: The height to set.
        """
        self._title_height = height
        self._title_widget.set_fixed_height(height)
        self._panel_option.set_icon_size(QtCore.QSize(height, height))

    def title_height(self) -> int:
        """Get the height of the title widget.

        :return: The height of the title widget.
        """
        return self._title_height

    def add_widgets_by(self, data: Dict[str, Dict]) -> Dict[str, QtWidgets.QWidget]:
        """Add widgets to the panel.

        :param data: The data to add. The dict is of the form:

            .. code-block:: python

                {
                    "widget-name": {
                        "type": "Button",
                        "args": (),
                        "kwargs": {  # or "arguments" for backward compatibility
                            "key1": "value1",
                            "key2": "value2"
                        }
                    },
                }

            Possible types are: Button, SmallButton, MediumButton, LargeButton,
            ToggleButton, SmallToggleButton, MediumToggleButton, LargeToggleButton, ComboBox, FontComboBox,
            LineEdit, TextEdit, PlainTextEdit, Label, ProgressBar, SpinBox, DoubleSpinBox, DataEdit, TimeEdit,
            DateTimeEdit, TableWidget, TreeWidget, ListWidget, CalendarWidget, Separator, HorizontalSeparator,
            VerticalSeparator, Gallery.
        :return: A dictionary of the added widgets.
        """
        widgets = {}  # type: Dict[str, QtWidgets.QWidget]
        for key, widget_data in data.items():
            type = widget_data.pop("type", "").capitalize()
            method = getattr(self, f"add{type}", None)  # type: Callable
            assert callable(method), f"Method add{type} is not callable or does not exist"
            args = widget_data.get("args", ())
            kwargs = widget_data.get("kwargs", widget_data.get("arguments", {}))
            widgets[key] = method(*args, **kwargs)
        return widgets

    def add_widget(
        self,
        widget: QtWidgets.QWidget,
        *,
        row_span: Union[int, RibbonButtonStyle] = Small,
        col_span: int = 1,
        mode: RibbonSpaceFindMode = ColumnWise,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag.AlignCenter,
        fixed_height: Union[bool, float] = False,
    ) -> QtWidgets.QWidget | Any:
        """Add a widget to the panel.

        :param widget: The widget to add.
        :param row_span: The number of rows the widget should span, 2: small, 3: medium, 6: large.
        :param col_span: The number of columns the widget should span.
        :param mode: The mode to find spaces.
        :param alignment: The alignment of the widget.
        :param fixed_height: Whether to fix the height of the widget, it can be a boolean, a percentage or a fixed
                            height, when a boolean is given, the height is fixed to the maximum height allowed if the
                            value is True, when a percentage is given (0 < percentage < 1) the height is calculated
                            from the height of the maximum height allowed, depends on the number of rows to span. The
                            minimum height is 40% of the maximum height allowed.
        :return: The added widget.
        """
        row_span = self.default_row_span(row_span)
        self._widgets.append(widget)
        row, col = self._grid_layout_manager.request_cells(row_span, col_span, mode)
        maximumHeight = self.row_height() * row_span + self._actions_layout.vertical_spacing() * (row_span - 2)
        widget.set_maximum_height(maximumHeight)
        if fixed_height is True or fixed_height > 0:
            fixed_height = (
                int(fixed_height * maximumHeight)
                if 0 < fixed_height <= 1
                else fixed_height if 1 < fixed_height < maximumHeight else maximumHeight
            )
            fixed_height = max(fixed_height, 0.4 * maximumHeight)  # minimum height is 40% of the maximum height
            widget.set_fixed_height(fixed_height)
        item = RibbonPanelItemWidget(self)
        item.add_widget(widget)
        self._actions_layout.add_widget(item, row, col, row_span, col_span, alignment)  # type: ignore
        return widget

    add_small_widget = functools.partialmethod(add_widget, row_span=Small)
    add_medium_widget = functools.partialmethod(add_widget, row_span=Medium)
    add_large_widget = functools.partialmethod(add_widget, row_span=Large)

    def remove_widget(self, widget: QtWidgets.QWidget):
        """Remove a widget from the panel."""
        self._actions_layout.remove_widget(widget)

    def widget(self, index: int) -> QtWidgets.QWidget:
        """Get the widget at the given index.

        :param index: The index of the widget, starting from 0.
        :return: The widget at the given index.
        """
        return self._widgets[index]

    def widgets(self) -> List[QtWidgets.QWidget]:
        """Get all the widgets in the panel.

        :return: A list of all the widgets in the panel.
        """
        return self._widgets

    def add_button(
        self,
        text: str = None,
        icon: QtGui.QIcon = None,
        show_text: bool = True,
        slot: Callable = None,
        shortcut: (
            QtCore.Qt.Key | QtGui.QKeySequence | QtCore.QKeyCombination | QtGui.QKeySequence.StandardKey | str | int
        ) = None,
        tooltip: str = None,
        statusTip: str = None,
        checkable: bool = False,
        *,
        row_span: RibbonButtonStyle = Large,
        **kwargs,
    ) -> RibbonToolButton:
        """Add a button to the panel.

        :param text: The text of the button.
        :param icon: The icon of the button.
        :param show_text: Whether to show the text of the button.
        :param slot: The slot to call when the button is clicked.
        :param shortcut: The shortcut of the button.
        :param tooltip: The tooltip of the button.
        :param statusTip: The status tip of the button.
        :param checkable: Whether the button is checkable.
        :param row_span: The type of the button corresponding to the number of rows it should span.
        :param kwargs: keyword arguments to control the properties of the widget on the ribbon bar.

        :return: The button that was added.
        """
        assert isinstance(row_span, RibbonButtonStyle), "row_span must be an instance of RibbonButtonStyle"
        style = row_span
        button = RibbonToolButton(self)
        button.set_button_style(style)
        button.set_text(text) if text else None
        button.set_icon(icon) if icon else None
        button.clicked.connect(slot) if slot else None  # type: ignore
        button.set_shortcut(shortcut) if shortcut else None
        button.set_tool_tip(tooltip) if tooltip else None
        button.set_status_tip(statusTip) if statusTip else None
        maximumHeight = (
            self.height()
            - self._title_label.size_hint().height()
            - self._main_layout.spacing()
            - self._main_layout.contents_margins().top()
            - self._main_layout.contents_margins().bottom()
        )
        button.set_maximum_height(maximumHeight)
        if style == Large:
            font_size = max(button.font().point_size() * 4 / 3, button.font().pixel_size())
            arrow_size = font_size
            maximum_icon_size = max(maximumHeight - font_size * 2 - arrow_size, 48)
            button.set_maximum_icon_size(int(maximum_icon_size))
        if not show_text:
            button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        button.set_checkable(checkable)
        kwargs["row_span"] = (
            self.default_row_span(Small)
            if style == Small
            else self.default_row_span(Medium) if style == Medium else self.default_row_span(Large)
        )
        self.add_widget(button, **kwargs)  # noqa
        return button

    add_small_button = functools.partialmethod(add_button, row_span=Small)
    add_medium_button = functools.partialmethod(add_button, row_span=Medium)
    add_large_button = functools.partialmethod(add_button, row_span=Large)
    add_toggle_button = functools.partialmethod(add_button, checkable=True)
    add_small_toggle_button = functools.partialmethod(add_toggle_button, row_span=Small)
    add_medium_toggle_button = functools.partialmethod(add_toggle_button, row_span=Medium)
    add_large_toggle_button = functools.partialmethod(add_toggle_button, row_span=Large)

    def add_ribbon_widget(
        self,
        *args,
        cls,
        initializer: Callable = None,
        row_span: Union[int, RibbonButtonStyle] = Small,
        col_span: int = 1,
        mode: RibbonSpaceFindMode = ColumnWise,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag.AlignCenter,
        fixed_height: Union[bool, float] = False,
        **kwargs,
    ) -> QtWidgets.QWidget:
        """Add any widget to the panel.

        :param cls: The class of the widget to add.
        :param initializer: The initializer function of the widget to add.
        :param args: The arguments passed to the initializer.
        :param row_span: The number of rows the widget should span, 2: small, 3: medium, 6: large.
        :param col_span: The number of columns the widget should span.
        :param mode: The mode to find spaces.
        :param alignment: The alignment of the widget.
        :param fixed_height: Whether to fix the height of the widget, it can be a boolean, a percentage or a fixed
                            height, when a boolean is given, the height is fixed to the maximum height allowed if the
                            value is True, when a percentage is given (0 < percentage < 1) the height is calculated
                            from the height of the maximum height allowed, depends on the number of rows to span. The
                            minimum height is 40% of the maximum height allowed.
        :param kwargs: The keyword arguments are passed to the initializer
        """
        widget = cls(self)
        if callable(initializer):
            initializer(widget, *args, **kwargs)
        elif args or kwargs:
            raise ValueError("Arguments are provided but the initializer is not set")
        return self.add_widget(
            widget=widget,
            row_span=row_span,
            col_span=col_span,
            mode=mode,
            alignment=alignment,
            fixed_height=fixed_height
        )

    def __getattr__(self, method: str) -> Callable:
        """Get the dynamic method `add[Small|Medium|Large][Widget]`.

        :param method: The name of the method to get.
        :return: The method of the widget to add.
        """
        # Match the method name
        match = re.match(r"add(Small|Medium|Large)(\w+)", method)
        assert match, f"Invalid method name: {method!s}"

        # Get the widget class and the size
        size = match.group(1)
        base_method_name = f"add{match.group(2)}"
        assert hasattr(self, base_method_name), f"Invalid method name {base_method_name}"

        # Get the base method
        base_method = getattr(self, base_method_name)
        row_span = Small if size == "Small" else Medium if size == "Medium" else Large

        # Create the new method
        return functools.partial(base_method, row_span=row_span)

    add_check_box = functools.partialmethod(
        add_ribbon_widget, cls=QtWidgets.QCheckBox, initializer=QtWidgets.QCheckBox.set_text
    )
    add_combo_box = functools.partialmethod(
        add_ribbon_widget, cls=QtWidgets.QComboBox, initializer=QtWidgets.QComboBox.add_items
    )
    add_line_edit = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QLineEdit)
    add_text_edit = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QTextEdit)
    add_plain_text_edit = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QPlainTextEdit)
    add_label = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QLabel, initializer=QtWidgets.QLabel.set_text)
    add_progress_bar = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QProgressBar)
    add_slider = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QSlider)
    add_spin_box = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QSpinBox)
    add_double_spin_box = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QDoubleSpinBox)
    add_date_edit = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QDateEdit)
    add_time_edit = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QTimeEdit)
    add_date_time_edit = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QDateTimeEdit)
    add_table_widget = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QTableWidget, row_span=Large)
    add_tree_widget = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QTreeWidget, row_span=Large)
    add_list_widget = functools.partialmethod(
        add_ribbon_widget, cls=QtWidgets.QListWidget, initializer=QtWidgets.QListWidget.add_items, row_span=Large
    )
    add_calendar_widget = functools.partialmethod(add_ribbon_widget, cls=QtWidgets.QCalendarWidget, row_span=Large)
    add_font_combo_box = functools.partialmethod(
        add_ribbon_widget,
        cls=QtWidgets.QFontComboBox,
        initializer=QtWidgets.QFontComboBox.add_items,
        row_span=Large
    )
    add_small_font_combo_box = functools.partialmethod(add_font_combo_box, row_span=Small)
    add_medium_font_combo_box = functools.partialmethod(add_font_combo_box, row_span=Medium)
    add_large_font_combo_box = functools.partialmethod(add_font_combo_box, row_span=Large)

    def add_separator(self, orientation=QtCore.Qt.Orientation.Vertical, width=6, **kwargs) -> RibbonSeparator:
        """Add a separator to the panel.

        :param orientation: The orientation of the separator.
        :param width: The width of the separator.
        :param kwargs: keyword arguments to control the properties of the widget on the ribbon bar.

        :return: The separator.
        """
        kwargs["row_span"] = Large if "row_span" not in kwargs else kwargs["row_span"]
        return self.add_widget(RibbonSeparator(orientation, width), **kwargs)

    add_horizontal_separator = functools.partialmethod(add_separator, orientation=QtCore.Qt.Orientation.Horizontal)
    add_vertical_separator = functools.partialmethod(add_separator, orientation=QtCore.Qt.Orientation.Vertical)

    def add_gallery(self, minimum_width: int = 800, popup_hide_on_click: bool = False, **kwargs) -> RibbonGallery:
        """Add a gallery to the panel.

        :param minimum_width: The minimum width of the gallery.
        :param popup_hide_on_click: Whether the gallery popup should be hidden when a user clicks on it.
        :param kwargs: keyword arguments to control the properties of the widget on the ribbon bar.

        :return: The gallery.
        """
        kwargs["row_span"] = Large if "row_span" not in kwargs else kwargs["row_span"]
        row_span = self.default_row_span(kwargs["row_span"])
        gallery = RibbonGallery(minimum_width, popup_hide_on_click, self)
        maximum_height = self.row_height() * row_span + self._actions_layout.vertical_spacing() * (row_span - 2)
        gallery.set_fixed_height(maximum_height)
        return self.add_widget(gallery, **kwargs)



from __feature__ import snake_case

import typing

from PySide6 import QtCore, QtGui, QtWidgets

from snakeribbon.constants import RibbonCategoryStyle, RibbonIcon
from snakeribbon.panel import RibbonPanel
from snakeribbon.separator import RibbonSeparator
from snakeribbon.utils import DataFile

if typing.TYPE_CHECKING:
    from .ribbonbar import RibbonBar  # noqa: F401


class RibbonCategoryLayoutButton(QtWidgets.QToolButton):
    """Previous/Next buttons in the category when the
    size is not enough for the widgets.
    """


class RibbonCategoryScrollArea(QtWidgets.QScrollArea):
    """Scroll area for the gallery"""


class RibbonCategoryScrollAreaContents(QtWidgets.QFrame):
    """Scroll area contents for the gallery"""


class RibbonCategoryLayoutWidget(QtWidgets.QFrame):
    """The category layout widget's category scroll area to arrange the widgets in the category."""

    display_options_button_clicked = QtCore.Signal()

    def __init__(self, parent=None):
        """Create a new category layout widget.

        :param parent: The parent widget.
        """
        super().__init__(parent)

        # Contents of the category scroll area
        self._category_scroll_area_contents = RibbonCategoryScrollAreaContents()  # type: ignore
        self._category_scroll_area_contents.set_size_policy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding
        )
        self._category_layout = QtWidgets.QHBoxLayout(self._category_scroll_area_contents)
        self._category_layout.set_contents_margins(0, 0, 0, 0)
        self._category_layout.set_spacing(0)
        self._category_layout.set_size_constraint(QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)

        # Category scroll area
        self._category_scroll_area = RibbonCategoryScrollArea()  # type: ignore
        self._category_scroll_area.set_horizontal_scroll_bar_policy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._category_scroll_area.set_vertical_scroll_bar_policy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._category_scroll_area.set_widget(self._category_scroll_area_contents)

        # Previous/Next buttons
        self._previous_button = RibbonCategoryLayoutButton(self)
        self._previous_button.set_icon(DataFile(RibbonIcon.Backward))
        self._previous_button.set_icon_size(QtCore.QSize(12, 12))
        self._previous_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._previous_button.set_auto_raise(True)
        self._previous_button.clicked.connect(self.scroll_previous)  # type: ignore
        self._next_button = RibbonCategoryLayoutButton(self)
        self._next_button.set_icon(DataFile(RibbonIcon.Forward))
        self._next_button.set_icon_size(QtCore.QSize(12, 12))
        self._next_button.set_tool_button_style(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._next_button.set_auto_raise(True)
        self._next_button.clicked.connect(self.scroll_next)  # type: ignore

        # Add the widgets to the main layout
        self._main_layout = QtWidgets.QHBoxLayout(self)
        self._main_layout.set_contents_margins(0, 0, 0, 0)
        self._main_layout.set_spacing(0)
        self._main_layout.add_widget(self._previous_button, 0, QtCore.Qt.AlignmentFlag.AlignVCenter)
        self._main_layout.add_widget(self._category_scroll_area, 1)
        self._main_layout.add_spacer_item(QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Policy.Expanding,
                                                               QtWidgets.QSizePolicy.Policy.Minimum))  # fmt: skip
        self._main_layout.add_widget(self._next_button, 0, QtCore.Qt.AlignmentFlag.AlignVCenter)

        # Auto set the visibility of the scroll buttons
        self.auto_set_scroll_buttons_visible()

    def paint_event(self, a0: QtGui.QPaintEvent) -> None:
        """Override the paint event to draw the background."""
        super().paint_event(a0)
        self.auto_set_scroll_buttons_visible()

    def resize_event(self, a0: QtGui.QResizeEvent) -> None:
        """Override the resize event to resize the scroll area."""
        super().resize_event(a0)
        self.auto_set_scroll_buttons_visible()

    def auto_set_scroll_buttons_visible(self):
        """Set the visibility of the scroll buttons."""
        horizontal_scroll_bar = self._category_scroll_area.horizontal_scroll_bar()
        self._previous_button.set_visible(horizontal_scroll_bar.value() > horizontal_scroll_bar.minimum())
        self._next_button.set_visible(horizontal_scroll_bar.value() < horizontal_scroll_bar.maximum())
        self._previous_button.set_icon_size(QtCore.QSize(12, self.size().height() - 15))
        self._next_button.set_icon_size(QtCore.QSize(12, self.size().height() - 15))

    def scroll_previous(self):
        """Scroll the category to the previous widget."""
        horizontal_scroll_bar = self._category_scroll_area.horizontal_scroll_bar()
        horizontal_scroll_bar.set_value(horizontal_scroll_bar.value() - 50)
        self.auto_set_scroll_buttons_visible()

    def scroll_next(self):
        """Scroll the category to the next widget."""
        self._category_scroll_area.horizontal_scroll_bar().set_value(
            self._category_scroll_area.horizontal_scroll_bar().value() + 50
        )
        self.auto_set_scroll_buttons_visible()

    def add_widget(self, widget: QtWidgets.QWidget):
        """Add a widget to the category layout.

        :param widget: The widget to add.
        """
        self._category_layout.add_widget(widget)

    def remove_widget(self, widget: QtWidgets.QWidget):
        """Remove a widget from the category layout.

        :param widget: The widget to remove.
        """
        self._category_layout.remove_widget(widget)

    def take_widget(self, widget: QtWidgets.QWidget) -> QtWidgets.QWidget:
        """Remove and return a widget from the category layout.

        :param widget: The widget to remove.
        :return: The widget that was removed.
        """
        self._category_layout.remove_widget(widget)
        return widget


class RibbonCategory(RibbonCategoryLayoutWidget):
    """The RibbonCategory is the logical grouping that represents the contents of a ribbon tab."""

    #: Title of the category
    _title: str
    #: The button style of the category.
    _style: RibbonCategoryStyle
    #: Panels
    _panels: typing.Dict[str, RibbonPanel]
    #: color of the context category
    _color: typing.Optional[QtGui.QColor]
    #: Maximum rows
    _max_rows: int = 6

    @typing.overload
    def __init__(
        self,
        title: str = "",
        style: RibbonCategoryStyle = RibbonCategoryStyle.Normal,
        color: QtGui.QColor = None,
        parent=None,
    ):
        pass

    @typing.overload
    def __init__(self, parent=None):
        pass

    def __init__(self, *args, **kwargs):
        """Create a new category.

        :param title: The title of the category.
        :param style: The button style of the category.
        :param color: The color of the context category.
        :param parent: The parent widget.
        """
        if (args and not isinstance(args[0], QtWidgets.QWidget)) or (
                "title" in kwargs or "style" in kwargs or "color" in kwargs
        ):
            title = args[0] if len(args) > 0 else kwargs.get("title", "")
            style = args[1] if len(args) > 1 else kwargs.get("style", RibbonCategoryStyle.Normal)
            color = args[2] if len(args) > 2 else kwargs.get("color", None)
            parent = args[3] if len(args) > 3 else kwargs.get("parent", None)
        else:
            title = ""
            style = RibbonCategoryStyle.Normal
            color = None
            parent = args[0] if len(args) > 0 else kwargs.get("parent", None)
        super().__init__(parent)
        self._title = title
        self._style = style
        self._panels = {}
        self._ribbon = parent  # type: RibbonBar
        self._color = color

    def set_maximum_rows(self, rows: int):
        """Set the maximum number of rows.

        :param rows: The maximum number of rows.
        """
        self._max_rows = rows

    def title(self) -> str:
        """Return the title of the category."""
        return self._title

    def set_category_style(self, style: RibbonCategoryStyle):
        """Set the button style of the category.

        :param style: The button style.
        """
        self._style = style
        self.repaint()

    def category_style(self) -> RibbonCategoryStyle:
        """Return the button style of the category.

        :return: The button style.
        """
        return self._style

    def add_panels_by(
        self,
        data: typing.Dict[
            str,  # title of the panel
            typing.Dict,  # data of the panel
        ],
    ) -> typing.Dict[str, RibbonPanel]:
        """Add panels from a dictionary.

        :param data: The dictionary. The keys are the titles of the panels. The value is a dictionary of
                     arguments. the argument showPanelOptionButton is a boolean to decide whether to show
                     the panel option button, the rest arguments are passed to the RibbonPanel.add_widgetsBy() method.
                     The dict is of the form:

                     .. code-block:: python

                        {
                            "panel-title": {
                                "showPanelOptionButton": True,
                                "widgets": {
                                    "widget-name": {
                                        "type": "Button",
                                        "args": (),
                                        "kwargs": {  # or "arguments" for backward compatibility
                                            "key1": "value1",
                                            "key2": "value2"
                                        }
                                    },
                                }
                            },
                        }
        :return: A dictionary of the newly created panels.
        """
        panels = {}
        for title, panel_data in data.items():
            show_panel_option_button = panel_data.get("show_panel_option_button", True)
            panels[title] = self.add_panel(title, show_panel_option_button)
            panels[title].add_widgets_by(panel_data.get("widgets", {}))
        return panels

    def add_panel(self, title: str, show_panel_option_button=True) -> RibbonPanel:
        """Add a new panel to the category.

        :param title: The title of the panel.
        :param show_panel_option_button: Whether to show the panel option button.
        :return: The newly created panel.
        """
        panel = RibbonPanel(title, max_rows=self._max_rows, show_panel_option_button=show_panel_option_button, parent=self)
        panel.set_fixed_height(
            self.height()
            - self._main_layout.spacing()
            - self._main_layout.contents_margins().top()
            - self._main_layout.contents_margins().bottom()
        )
        self._panels[title] = panel
        self.add_widget(panel)  # type: ignore
        self.add_widget(RibbonSeparator(width=10))  # type: ignore
        return panel

    def remove_panel(self, title: str):
        """Remove a panel from the category.

        :param title: The title of the panel.
        """
        # self._panelLayout.removeWidget(self._panels[title])
        self.remove_widget(self._panels[title])
        self._panels.pop(title)

    def take_panel(self, title: str) -> RibbonPanel:
        """Remove and return a panel from the category.

        :param title: The title of the panel.
        :return: The removed panel.
        """
        panel = self._panels[title]
        self.remove_panel(title)
        return panel

    def panel(self, title: str) -> RibbonPanel:
        """Return a panel from the category.

        :param title: The title of the panel.
        :return: The panel.
        """
        return self._panels[title]

    def panels(self) -> typing.Dict[str, RibbonPanel]:
        """Return all panels in the category.

        :return: The panels.
        """
        return self._panels


class RibbonNormalCategory(RibbonCategory):
    """A normal category."""

    def __init__(self, title: str, parent: QtWidgets.QWidget):
        """Create a new normal category.

        :param title: The title of the category.
        :param parent: The parent widget.
        """
        super().__init__(title, RibbonCategoryStyle.Normal, parent=parent)

    def set_category_style(self, style: RibbonCategoryStyle):
        """Set the button style of the category.

        :param style: The button style.
        """
        raise ValueError("You can not set the category style of a normal category.")


class RibbonContextCategory(RibbonCategory):
    """A context category."""

    def __init__(self, title: str, color: QtGui.QColor, parent: QtWidgets.QWidget):
        """Create a new context category.

        :param title: The title of the category.
        :param color: The color of the context category.
        :param parent: The parent widget.
        """
        super().__init__(title, RibbonCategoryStyle.Context, color=color, parent=parent)

    def set_category_style(self, style: RibbonCategoryStyle):
        """Set the button style of the category.

        :param style: The button style.
        """
        raise ValueError("You can not set the category style of a context category.")

    def color(self) -> QtGui.QColor:
        """Return the color of the context category.

        :return: The color of the context category.
        """
        return self._color

    def set_color(self, color: QtGui.QColor):
        """Set the color of the context category.

        :param color: The color of the context category.
        """
        self._color = color
        self._ribbon.repaint()

    def show_context_category(self):
        """Show the given category, if it is not a context category, nothing happens."""
        self._ribbon.show_context_category(self)

    def hide_context_category(self):
        """Hide the given category, if it is not a context category, nothing happens."""
        self._ribbon.hide_context_category(self)

    def category_visible(self) -> bool:
        """Return whether the category is shown.

        :return: Whether the category is shown.
        """
        return self._ribbon.category_visible(self)

    def set_category_visible(self, visible: bool):
        """Set the state of the category.

        :param visible: The state.
        """
        if visible:
            self.show_context_category()
        else:
            self.hide_context_category()


class RibbonContextCategories(typing.Dict[str, RibbonContextCategory]):
    """A list of context categories."""

    def __init__(
        self,
        name: str,
        color: QtGui.QColor,
        categories: typing.Dict[str, RibbonContextCategory],
        ribbon,
    ):
        self._name = name
        self._color = color
        self._ribbon = ribbon
        super().__init__(categories)

    def name(self) -> str:
        """Return the name of the context categories."""
        return self._name

    def set_name(self, name: str):
        """Set the name of the context categories."""
        self._name = name

    def color(self) -> QtGui.QColor:
        """Return the color of the context categories."""
        return self._color

    def set_color(self, color: QtGui.QColor):
        """Set the color of the context categories."""
        self._color = color

    def show_context_categories(self):
        """Show the categories"""
        self._ribbon.show_context_category(self)

    def hide_context_categories(self):
        """Hide the categories"""
        self._ribbon.hide_context_category(self)

    def categories_visible(self) -> bool:
        """Return whether the categories are shown."""
        for category in self.values():
            if category.category_visible():
                return True
        return False

    def set_categories_visible(self, visible: bool):
        """Set the state of the categories."""
        self.show_context_categories() if visible else self.hide_context_categories()
